# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from kubernetes import client as k8sClient
from azdata.cli.core.exceptions import (KubernetesError)
from kubernetes.client.rest import ApiException as K8sApiException
from azdata.cli.core.logging import get_logger

logger = get_logger(__name__)


class KubernetesClient(object):
    """
    Client for managing Kubernetes Resources, to be used by Arc partner command modules.
    """

    """
    Custom Resource Definition specification key.
    """
    CUSTOM_RESOURCE_DEFINITION = 'CustomResourceDefinition'

    @staticmethod
    def get_service(ns, service_name):
        """
        Get the specified service in the given namespace.
        :param ns:
        :return:
        """
        try:
            return k8sClient.CoreV1Api().read_namespaced_service(service_name, ns)
        except K8sApiException as e:
            logger.error(e.body)
            raise

    @staticmethod
    def service_exists(ns, service_name):
        """
        Get the specified service in the given namespace.
        :param ns:
        :return:
        """
        try:
            service = k8sClient.CoreV1Api().read_namespaced_service(service_name, ns)
            return service_name == service.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def list_pods(ns, label_selector=None):
        """
        List the pods in the given namespace
        :param ns: The namespace
        :param label_selector: The label to select the pods by.
        :return:
        """
        try:
            return k8sClient.CoreV1Api().list_namespaced_pod(ns, label_selector=label_selector)
        except K8sApiException as e:
            logger.error(e.body)
            raise

    @staticmethod
    def create_secret(ns, config):
        """
        Creates a secret in Kubernetes
        :param ns:
        :param config:
        :return:
        """
        try:
            k8sClient.CoreV1Api().create_namespaced_secret(namespace=ns, body=config)
        except K8sApiException as e:
            logger.error(e.body)
            raise

    @staticmethod
    def secret_exists(ns, secret_name):
        """
        Returns true if secret exists in the namespace
        :param ns:
        :param secret_name:
        :return:
        """
        try:
            secret = k8sClient.CoreV1Api().read_namespaced_secret(secret_name, ns)
            return secret_name == secret.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def create_replica_set(ns, spec):
        """
        Creates a replica set in the given namespace
        :param ns:
        :param spec:
        :return:
        """
        try:
            k8sClient.AppsV1Api().create_namespaced_replica_set(ns, spec)
        except K8sApiException as e:
            logger.debug(e.body)
            raise e

    @staticmethod
    def replica_set_exists(ns, replica_set_name):
        """
        Returns true if replica set exists in the namespace
        :param ns:
        :param replica_set_name:
        :return:
        """
        try:
            replica_set = k8sClient.AppsV1Api().read_namespaced_replica_set(replica_set_name, ns)
            return replica_set_name == replica_set.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def service_account_exists(ns, service_account_name):
        """
        Returns true if service account exists in the namespace
        :param ns:
        :param service_account_name:
        :return:
        """
        try:
            service_account = k8sClient.CoreV1Api().read_namespaced_service_account(service_account_name, ns)
            return service_account_name == service_account.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def namespaced_role_exists(ns, role_name):
        """
        Returns true if role exists in the namespace
        :param ns:
        :param role_name:
        :return:
        """
        try:
            role = k8sClient.RbacAuthorizationV1Api().read_namespaced_role(role_name, ns)
            return role_name == role.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    @staticmethod
    def namespaced_role_binding_exists(ns, role_binding_name):
        """
        Returns true if role binding exists in the namespace
        :param ns:
        :param role_binding_name:
        :return:
        """
        try:
            role_binding = k8sClient.RbacAuthorizationV1Api().read_namespaced_role_binding(role_binding_name,
                                                                                           ns)
            return role_binding_name == role_binding.metadata.name
        except K8sApiException as e:
            logger.debug(e.body)
            return False

    # ------------------------------
    # Custom Resource Methods
    # ------------------------------

    @staticmethod
    def create_custom_resource_definition(crd):
        """
        Create the given crd.
        :param crd: The crd to create.
        :return:
        """
        try:
            api = k8sClient.ApiextensionsV1beta1Api()
            current_crds = [x['spec']['names']['kind'].lower()
                            for x in api.list_custom_resource_definition().to_dict()['items']]
            if crd.kind.lower() not in current_crds:
                return api.create_custom_resource_definition(crd.body)

        except ValueError as e:
            # Kubernetes API bug with prior versions of Kubernetes
            # The CRD does get created but returns an erroneous value error that we can safely ignore.
            # This is fixed in Kubernetes version 1.16
            # https://github.com/kubernetes-client/python/issues/1022
            # We can remove this condition when we upgrade to version 1.16, server-side.
            if 'Invalid value for `conditions`, must not be `None`' in str(e):
                return
            else:
                raise e
        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def create_namespaced_custom_object(crd, cr):
        """
        Creates a kubernetes custom resource object with the given crd and specification json
        :param crd: The definition of the custom resource to be deployed.
        :param cr: The custom resource instance to be deployed.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.create_namespaced_custom_object(
                body=cr.body,
                namespace=cr.namespace,
                plural=crd.plural,
                group=cr.group,
                version=cr.version
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def patch_namespaced_custom_object(crd, cr):
        """
        Patches a kubernetes custom resource object with the given crd and specification json
        :param crd: The definition of the custom resource to be deployed.
        :param cr: The custom resource instance to be deployed.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.patch_namespaced_custom_object(
                body=cr.body,
                namespace=cr.namespace,
                plural=crd.plural,
                group=cr.group,
                version=cr.version,
                name=cr.name
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def delete_namespaced_custom_object(crd, name, namespace):
        """
        Deletes a custom resource object
        :param crd: The definition of the custom resource to be deleted.
        :param name: The name of the custom resource to be deleted.
        :param namespace: The namespace from which to delete the custom resource.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.delete_namespaced_custom_object(
                group=crd.group,
                version=crd.version,
                plural=crd.plural,
                namespace=namespace,
                name=name,
                body=k8sClient.V1DeleteOptions()
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def get_namespaced_custom_object(crd, name, namespace):
        """
        Gets the custom resource object
        :param crd: The definition of the custom resource.
        :param name: The name of the custom resource.
        :param namespace: The namespace of the custom resource.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.get_namespaced_custom_object(
                group=crd.group,
                version=crd.version,
                name=name,
                namespace=namespace,
                plural=crd.plural,
            )

        except Exception as e:
            raise KubernetesError(e)

    @staticmethod
    def list_namespaced_custom_object(crd, namespace):
        """
        Lists the custom resource object
        :param crd: The definition of the custom resource.
        :param namespace: The namespace of the custom resource.
        :return:
        """
        try:
            api = k8sClient.CustomObjectsApi()
            return api.list_namespaced_custom_object(
                group=crd.group,
                version=crd.version,
                namespace=namespace,
                plural=crd.plural,
            )

        except Exception as e:
            raise KubernetesError(e)


class CustomResource(object):
    """
    Internal Custom Resource object to be used for deployments.
    """
    def __init__(self, cr_object):
        """
        Initializes a CR object with the given json.
        :param cr_object: The custom resource json.
        """
        self.body = cr_object

    @property
    def body(self):
        """
        Gets the cr object used for deployments.
        """
        return self._body

    @body.setter
    def body(self, cr_body):
        """
        Sets the cr json body used for deployments.
        :param cr_body: The custom resource json body.
        """
        self._body = cr_body

    @property
    def metadata(self):
        """
        Returns the metadata of the cr.
        :return:
        """
        return self.body['metadata']

    @property
    def name(self):
        """
        Returns the name of the cr.
        :return:
        """
        return self.metadata['name']

    @name.setter
    def name(self, n):
        """
        Sets the name of the cr.
        :param n: Name of the cr.
        :return:
        """
        self.metadata['name'] = n

    @property
    def namespace(self):
        """
        Returns the namespace of the cr.
        :return:
        """
        return self.metadata['namespace']

    @namespace.setter
    def namespace(self, ns):
        """
        Sets the namespace of the cr.
        :param ns: Namespace of the cr.
        :return:
        """
        self.metadata['namespace'] = ns

    @property
    def api(self):
        """
        Returns the api information of the cr.
        :return:
        """
        return self.body['apiVersion']

    @property
    def group(self):
        """
        Returns the api group of the cr.
        :return:
        """
        return self.api.split('/')[0]

    @property
    def version(self):
        """
        Returns the api version of the cr.
        :return:
        """
        return self.api.split('/')[1]


class CustomResourceDefinition(object):
    """
    Internal Custom Resource Definition object to be used for deployments.
    """
    def __init__(self, crd_object):
        """
        Initializes a CRD object with the given json.
        :param crd_object:
        """
        self.body = crd_object

    @property
    def body(self):
        """
        Gets the crd json body used for deployments.
        """
        return self._body

    @body.setter
    def body(self, crd_body):
        """
        Sets the crd body used for deployments.
        :param crd_body: The crd json.
        """
        self._body = crd_body

    @property
    def metadata(self):
        """
        Returns the metadata of the crd.
        :return:
        """
        return self.body['metadata']

    @property
    def name(self):
        """
        Returns the name of the crd.
        :return:
        """
        return self.metadata['name']

    @property
    def kind(self):
        """
        Returns the kind of the crd.
        :return:
        """
        return self.names['kind']

    @property
    def spec(self):
        """
        Gets the spec of this crd.
        :return:
        """
        return self.body['spec']

    @property
    def group(self):
        """
        Gets the API group of this CRD
        :return:
        """
        return self.spec['group']

    @property
    def version(self):
        """
        Gets the API version of this CRD
        :return:
        """
        return self.spec['version']

    @property
    def names(self):
        """
        Gets the names of this crd.
        :return:
        """
        return self.spec['names']

    @property
    def plural(self):
        """
        Gets the plural form of this crd.
        :return:
        """
        return self.names['plural']

